#include "Terrain/ClipmapSampling.hlsl"

// Input height maps
Texture2D heightTexture;
Texture2D coarseHeightTexture;

SamplerState heightSampler;
SamplerState coarseHeightSampler;

float2 convertTexCoords(float2 texCoords)
{
#ifdef COGS_SHADER_GL
  return texCoords;
#else
  return float2(0, 1) + float2(1, -1) * texCoords;
#endif
}

float4 calculateTerrainTexCoords(in float2 levelPos)
{
  float2 fineTexCoords = convertTexCoords((levelPos + fineTextureOrigin) * oneOverClipmapSize);
  float2 coarseTexCoords = convertTexCoords((levelPos * fineToCoarseRatio + fineLevelOriginInCoarse) * oneOverCoarseClipmapSize);

  return float4(fineTexCoords, coarseTexCoords);
}

float calculateBlendFactor(in float2 levelPos)
{
  if (!useBlendRegions) return 0.0;

  // Apply 5% adjustment factor to increase the region of fully blended geometry.
  const float adjustment = 1.05;

  float2 blendFactor = saturate((abs(levelPos - viewPosInClippedLevel) - unblendedRegionSize) * oneOverBlendedRegionSize * adjustment);

  return saturate(max(blendFactor.x, blendFactor.y));
}

// Samples the height of the current clipmap level at the given position (XY range from [0, 1]).
float sampleHeight(in float4 texCoords, in float blendFactor)
{
#ifdef GLSLANG
  texCoords.yw = (float2)1.0 - texCoords.yw;
#endif
  float fineHeight = heightTexture.SampleLevel(heightSampler, texCoords.xy, 0).r;
  float coarseHeight = coarseHeightTexture.SampleLevel(coarseHeightSampler, texCoords.zw, 0).r;

  return lerp(fineHeight, coarseHeight, blendFactor);
}

float sampleHeightNoData(in float4 texCoords, in float blendFactor, out float noDataFactor)
{
#ifdef GLSLANG
  texCoords.yw = (float2)1.0 - texCoords.yw;
#endif
  float fineHeight = heightTexture.SampleLevel(heightSampler, texCoords.xy, 0).r;
  float coarseHeight = sampleLinearNoData(coarseHeightTexture, coarseHeightSampler, texCoords.zw, noClipmapData);

  return lerpNoData(fineHeight, coarseHeight, blendFactor, noClipmapData, noDataFactor);
}

float sampleHeight(inout VertexOut Out, float2 levelPos)
{
  Out.TerrainTexCoords = calculateTerrainTexCoords(levelPos);
  Out.BlendFactor = calculateBlendFactor(levelPos);

  if (!enableNoData) {
    return sampleHeight(Out.TerrainTexCoords, Out.BlendFactor);
  } else {
    return sampleHeightNoData(Out.TerrainTexCoords, Out.BlendFactor, Out.NodataFactor);
  }
}

float applyClipping(float4 position)
{
  float clipDistance = 1;

  if (useClipping) {
    float middleX = (clipmapExtents.x + clipmapExtents.z) * 0.5;
    float middleY = (clipmapExtents.y + clipmapExtents.w) * 0.5;

    float clipX = 0;
    float clipY = 0;

    if (position.x > middleX) {
      clipX = position.x - clipmapExtents.z;
    } else {
      clipX = clipmapExtents.x - position.x;
    }

    if (position.y > middleY) {
      clipY = position.y - clipmapExtents.w;
    } else {
      clipY = clipmapExtents.y - position.y;
    }

    clipDistance = max(clipX, clipY);
  }

  return clipDistance;
}
