#include "Terrain/ConstantBuffers.hlsl"
#include "Terrain/ClipmapCommonVS.hlsl"
#include "Terrain/Noise.hlsl"
#include "Terrain/OceanCommon.hlsl"

float mySampleHeight(in Texture2D source, in float2 texCoords, in float2 offset, float normalFactor)
{
  return source.SampleLevel(coarseHeightSampler, texCoords + offset * normalFactor, 0).r;
}

float3 computeLevelNormal(in Texture2D source, in float2 texCoords, float2 offsetSize, float normalFactor)
{
  float t = mySampleHeight(source, texCoords, float2(0.0, offsetSize.y), normalFactor).r;
  float b = mySampleHeight(source, texCoords, float2(0.0, -offsetSize.y), normalFactor).r;
  float l = mySampleHeight(source, texCoords, float2(-offsetSize.x, 0.0), normalFactor).r;
  float r = mySampleHeight(source, texCoords, float2(offsetSize.x, 0.0), normalFactor).r;

  float dx = l - r;
  float dy = t - b;

  float2 invScale = float2(heightExaggeration, heightExaggeration) / (levelScaleFactor * levelZeroWorldScaleFactor * worldScale.xy);

  return normalize(float3(dx * invScale.x, dy * invScale.y, 1));
}

float3 computeNormal(in float2 levelPos, float normalFactor)
{
  float4 TerrainTexCoords;
  TerrainTexCoords.xy = convertTexCoords((levelPos + fineTextureOrigin) * oneOverClipmapSize);
  TerrainTexCoords.zw = convertTexCoords((levelPos * fineToCoarseRatio + fineLevelOriginInCoarse) * oneOverCoarseClipmapSize);

  float BlendFactor = 0.0;
  if (useBlendRegions) {
    float2 blendFactor = clamp((abs(levelPos - viewPosInClippedLevel) - unblendedRegionSize) * oneOverBlendedRegionSize, 0.0, 1.0);

    BlendFactor = clamp(max(blendFactor.x, blendFactor.y), 0, 1);
  }

  float3 fineNormal = computeLevelNormal(heightTexture, TerrainTexCoords.xy, oneOverClipmapSize, normalFactor);

  if (useBlendRegions && BlendFactor != 0) {
    float3 coarseNormal = computeLevelNormal(coarseHeightTexture, TerrainTexCoords.zw, oneOverCoarseClipmapSize * fineToCoarseRatio, normalFactor);

    return lerp(fineNormal, coarseNormal, BlendFactor);
  }
  else {
    return fineNormal;
  }
}

VertexOut vertexFunction(VertexIn In)
{
  VertexOut Out = (VertexOut)0;

  float2 levelPos = In.Position.xy + patchOriginInClippedLevel;
  float2 clipmapPos = (levelPos + levelOffsetFromClipmapCenter) * levelScaleFactor * levelZeroWorldScaleFactor;
  float2 worldPos = clipmapPos * worldScale.xy + centerOffsetFromOrigin;

#ifdef ENABLE_IMAGERY
  [unroll(NUM_ACTIVE_TEXTURES)]
  for (int i = 0; i < NUM_ACTIVE_TEXTURES; ++i) {
    float2 imageryLevelPos = levelPos * terrainToImageryResolutionRatio[i];
    Out.ImageryTexCoords[i].xy = convertTexCoords((imageryLevelPos + terrainOffsetInImagery[i]) * oneOverImagerySize[i]);
    Out.ImageryTexCoords[i].zw = convertTexCoords(((imageryLevelPos * fineToCoarseImageryResolutionRatio[i]) + fineImageryOriginInCoarse[i]) * oneOverCoarseImagerySize[i]);
  }
#endif

  // Add small offset to allow discard to work after clamping.
  float offset = 0.1;

  float x = clamp(worldPos.x, worldExtents.x - offset, worldExtents.z + offset) + worldOffset.x;
  float y = clamp(worldPos.y, worldExtents.y - offset, worldExtents.w + offset) + worldOffset.y;

  float depth = saturate(-sampleHeight(Out, levelPos) * heightExaggeration / shoale);
  if (depth < 1.0) {
    Out.Seabed = float4(computeNormal(levelPos, 1), depth);
    if (length(Out.Seabed.xy) < 0.01) {
      Out.Seabed = float4(0, 0, 0, 1);
    }
  } else {
    Out.Seabed = float4(0, 0, 0, 1);
  }

  float3 camY = inverseViewMatrix[1].xyz;

  float4 a0 = mul(float4(x, y, 0, 1), worldToClipMatrix);
  float4 a1 = mul(float4(float3(x, y, 0) + significantWaveHeightWorld*camY, 1), worldToClipMatrix);
  float2 h = viewportScale*((1.0 / a0.w)*a0.xy - (1.0 / a1.w)*a1.xy);

  float lod = length(h);

  Out.OceanPosition = float4(x, y, lod, 0);

  float3 displacedOcean = waveDisplace(float2(x, y), Out.Seabed.w, Out.Seabed.xyz, lod);

  Out.Position = mul(float4(displacedOcean, 1.0), worldToClipMatrix);
  Out.WorldPosition = float4(displacedOcean, 1.0);

  Out.ScreenPosition = Out.Position;

  Out.ClipDistance = applyClipping(Out.WorldPosition - worldOffset);

  Out.OceanFactors = float4(0, 0, 0, 0);

  return Out;
}
